
import asyncio, sys, os, time

import xml.etree.ElementTree as ET

import fcntl

import logging
logger = logging.getLogger(__name__)

# All xml data received from the client, or from snooped devices should be contained in one of the following tags
TAGS = (b'getProperties',
        b'newTextVector',
        b'newNumberVector',
        b'newSwitchVector',
        b'newBLOBVector',
        b'enableBLOB',
        # below are tags from snooped devices
        b'message',
        b'delProperty',
        b'defSwitchVector',
        b'setSwitchVector',
        b'defLightVector',
        b'setLightVector',
        b'defTextVector',
        b'setTextVector',
        b'defNumberVector',
        b'setNumberVector',
        b'defBLOBVector',
        b'setBLOBVector'
       )


# _STARTTAGS is a tuple of ( b'<defTextVector', ...  ) data received will be tested to start with such a starttag
_STARTTAGS = tuple(b'<' + tag for tag in TAGS)


# _ENDTAGS is a tuple of ( b'</defTextVector>', ...  ) data received will be tested to end with such an endtag
_ENDTAGS = tuple(b'</' + tag + b'>' for tag in TAGS)



def _makestart(element):
    "Given an xml element, returns a string of its start, including < tag attributes >"
    attriblist = ["<", element.tag]
    for key,value in element.attrib.items():
        attriblist.append(f" {key}=\"{value}\"")
    attriblist.append(">")
    return "".join(attriblist)


async def queueget(queue, timeout=0.5):
    """"Returns True, True if timed out
                True, False is reserved for future
                False, Value if a value is taken from the queue"""
    try:
        value = await asyncio.wait_for(queue.get(), timeout)
    except asyncio.TimeoutError:
        return True, True
    return False, value


class STDOUT_TX:
    "An object that transmits data on stdout, used by STDINOUT as one half of the communications path"

    def __init__(self):
        self._stop = False       # Gets set to True to stop communications

    def shutdown(self):
        self._stop = True

    @property
    def stop(self):
        "returns self._stop, being the instruction to stop the driver"
        return self._stop


    async def run_tx(self, writerque):
        """Gets data from writerque, and transmits it out on stdout"""
        while not self._stop:
            await asyncio.sleep(0)
            # get block of data from writerque and transmit down stdout
            quexit, txdata = await queueget(writerque)
            if quexit:
                continue
            writerque.task_done()
            if txdata is None:
                await asyncio.sleep(0.02)
                continue
            if (txdata.tag == "setBLOBVector") and len(txdata):
                # txdata is a setBLOBVector containing blobs
                # send initial setBLOBVector
                startdata = _makestart(txdata)
                sys.stdout.buffer.write(startdata.encode())
                sys.stdout.buffer.flush()
                for oneblob in txdata.iter('oneBLOB'):
                    bytescontent = oneblob.text.encode()
                    # send start of oneblob
                    startoneblob = _makestart(oneblob)
                    sys.stdout.buffer.write(startoneblob.encode())
                    sys.stdout.buffer.flush()
                    # send content in chunks
                    chunksize = 1000
                    for b in range(0, len(bytescontent), chunksize):
                        byteschunk = bytescontent[b:b+chunksize]
                        sys.stdout.buffer.write(byteschunk)
                        sys.stdout.buffer.flush()
                        await asyncio.sleep(0)
                    sys.stdout.buffer.write(b"</oneBLOB>")
                    sys.stdout.buffer.flush()
                # send enddata
                sys.stdout.buffer.write(b"</setBLOBVector>\n")
                sys.stdout.buffer.flush()
            else:
                # its straight xml, send it out on stdout
                binarydata = ET.tostring(txdata)
                binarydata += b"\n"
                sys.stdout.buffer.write(binarydata)
                sys.stdout.buffer.flush()


class STDIN_RX:
    """An object that receives data on stdin, parses it to ElementTree elements
       and passes it to the driver by appending it to the driver's readerque"""

    def __init__(self):
        self._remainder = b""    # Used to store intermediate data
        self._stop = False       # Gets set to True to stop communications

    def shutdown(self):
        self._stop = True

    @property
    def stop(self):
        "returns self._stop, being the instruction to stop the driver"
        return self._stop

    async def run_rx(self, readerque):
        "pass data to readerque"
        try:
            # get block of xml.etree.ElementTree data
            # from self._xmlinput and append it to readerque
            while not self._stop:
                rxdata = await self._xmlinput()
                if rxdata is None:
                    return
                # append it to readerque
                while not self._stop:
                    try:
                        await asyncio.wait_for(readerque.put(rxdata), timeout=0.5)
                    except asyncio.TimeoutError:
                        # queue is full, continue while loop, checking stop flag
                        continue
                    # rxdata is now in readerque, break the inner while loop
                    break
        except Exception:
            logger.exception("Exception report from STDIN_RX.run_rx")
            raise

    async def _xmlinput(self):
        """get data from  _datainput, parse it, and return it as xml.etree.ElementTree object
           Returns None if stop flags arises"""
        message = b''
        messagetagnumber = None
        while not self._stop:
            await asyncio.sleep(0)
            data = await self._datainput()
            # data is either None, or binary data ending in b">"
            if data is None:
                return
            if self._stop:
                return
            if not message:
                # data is expected to start with <tag, first strip any newlines
                data = data.strip()
                for index, st in enumerate(_STARTTAGS):
                    if data.startswith(st):
                        messagetagnumber = index
                        break
                    elif st in data:
                        # remove any data prior to a starttag
                        positionofst = data.index(st)
                        data = data[positionofst:]
                        messagetagnumber = index
                        break
                else:
                    # data does not start with a recognised tag, so ignore it
                    # and continue waiting for a valid message start
                    continue
                # set this data into the received message
                message = data
                # either further children of this tag are coming, or maybe its a single tag ending in "/>"
                if message.endswith(b'/>'):
                    # the message is complete, handle message here
                    try:
                        root = ET.fromstring(message.decode("us-ascii"))
                    except Exception as e:
                        # failed to parse the message, continue at beginning
                        message = b''
                        messagetagnumber = None
                        continue
                    # xml datablock done, return it
                    return root
                # and read either the next message, or the children of this tag
                continue
            # To reach this point, the message is in progress, with a messagetagnumber set
            # keep adding the received data to message, until an endtag is reached
            message += data
            if message.endswith(_ENDTAGS[messagetagnumber]):
                # the message is complete, handle message here
                try:
                    root = ET.fromstring(message.decode("us-ascii"))
                except Exception as e:
                    # failed to parse the message, continue at beginning
                    message = b''
                    messagetagnumber = None
                    continue
                # xml datablock done, return it
                return root
            # so message is in progress, with a messagetagnumber set
            # but no valid endtag received yet, so continue the loop


    async def _datainput(self):
        """Waits for binary string of data ending in > from stdin
           Returns None if stop flags arises"""
        remainder = self._remainder
        if b">" in remainder:
            # This returns with binary data ending in > as long
            # as there are > characters in self._remainder
            binarydata, self._remainder = remainder.split(b'>', maxsplit=1)
            binarydata += b">"
            return binarydata
        # As soon as there are no > characters left in self._remainder
        # get more data from stdin
        while not self._stop:
            await asyncio.sleep(0)
            indata = sys.stdin.buffer.read(100)
            if not indata:
                await asyncio.sleep(0.02)
                continue
            remainder += indata
            if b">" in indata:
                binarydata, self._remainder = remainder.split(b'>', maxsplit=1)
                binarydata += b">"
                return binarydata



class STDINOUT():
    """If indipydriver.comms is set to an instance of this class it is
       used to implement communications via stdin and stdout"""

    def __init__(self):
        self.connected = True
        self.rx = STDIN_RX()
        self.tx = STDOUT_TX()
        self._stop = False       # Gets set to True to stop communications


    @property
    def stop(self):
        "returns self._stop, being the instruction to stop the driver"
        return self._stop

    def shutdown(self):
        self._stop = True
        self.rx.shutdown()
        self.tx.shutdown()

    async def __call__(self, readerque, writerque):
        "Called from indipydriver.asyncrun() to run the communications"
        # Set stdin to non-blocking mode
        flags = fcntl.fcntl(sys.stdin.fileno(), fcntl.F_GETFL)
        fcntl.fcntl(sys.stdin.fileno(), fcntl.F_SETFL, flags | os.O_NONBLOCK)

        logger.info("Communicating via STDIN/STDOUT")

        await asyncio.gather(self.rx.run_rx(readerque),
                             self.tx.run_tx(writerque)
                             )

class Port_TX():
    "An object that transmits data on a port, used by Portcomms as one half of the communications path"

    def __init__(self, sendchecker, writer):
        self.sendchecker = sendchecker
        self.writer = writer
        self._stop = False       # Gets set to True to stop communications

    @property
    def stop(self):
        "returns self._stop, being the instruction to stop the driver"
        return self._stop

    def shutdown(self):
        self._stop = True

    async def run_tx(self, writerque):
        """Gets data from writerque, and transmits it out on the port writer"""
        while not self._stop:
            await asyncio.sleep(0)
            # get block of data from writerque and transmit
            quexit, txdata = await queueget(writerque)
            if quexit:
                continue
            writerque.task_done()
            if txdata is None:
                continue
            if not self.sendchecker.allowed(txdata):
                # this data should not be transmitted, discard it
                continue
            # this data can be transmitted
            binarydata = ET.tostring(txdata)
            # Send to the port
            self.writer.write(binarydata)
            await self.writer.drain()
        self.writer.close()
        await self.writer.wait_closed()


class Port_RX(STDIN_RX):
    """Produces xml.etree.ElementTree data from data received on the port,
       this is used by Portcomms as one half of the communications path.
       This overwrites methods of the STDIN_RX parent class."""

    def __init__(self, sendchecker, reader):
        super().__init__()
        self.sendchecker = sendchecker
        self.reader = reader


    async def run_rx(self, readerque):
        "pass xml.etree.ElementTree data to readerque"
        try:
            # get block of xml.etree.ElementTree data
            # from self._xmlinput and append it to  readerque
            while not self._stop:
                rxdata = await self._xmlinput()
                if rxdata is None:
                    return
                if rxdata.tag == "enableBLOB":
                    # set permission flags in the sendchecker object
                    self.sendchecker.setpermissions(rxdata)
                # and place rxdata into readerque
                while not self._stop:
                    try:
                        await asyncio.wait_for(readerque.put(rxdata), timeout=0.5)
                    except asyncio.TimeoutError:
                        # queue is full, continue while loop, checking stop flag
                        continue
                    # rxdata is now in readerque, break the inner while loop
                    break
        except ConnectionError:
            # re-raise this without creating a report, as it probably indicates
            # a normal connection drop
            raise
        except Exception:
            # possibly some other error, so report it
            logger.exception("Exception report from Port_RX.run_rx")
            raise


    async def _datainput(self):
        """Waits for binary string of data ending in > from the port
           Returns None if stop flags arises"""
        binarydata = b""
        while not self._stop:
            await asyncio.sleep(0)
            try:
                data = await self.reader.readuntil(separator=b'>')
            except asyncio.LimitOverrunError:
                data = await self.reader.read(n=32000)
            except asyncio.IncompleteReadError:
                binarydata = b""
                continue
            if not data:
                await asyncio.sleep(0.01)
                continue
            # data received
            if b">" in data:
                binarydata = binarydata + data
                return binarydata
            # data has content but no > found
            binarydata += data
            # could put a max value here to stop this increasing indefinetly


class Portcomms():
    """If indipydriver.comms is set to an instance of this class it is
       used to implement communications via a port"""

    def __init__(self, devices, host="localhost", port=7624):
        # devices is a dictionary of device name to device this driver owns
        self.devices = devices
        self.sendchecker = SendChecker(devices)
        self.host = host
        self.port = port
        self.connected = False

        self.rx = None
        self.tx = None
        self.server = None
        self._stop = False       # Gets set to True to stop communications

    @property
    def stop(self):
        "returns self._stop, being the instruction to stop the driver"
        return self._stop

    def shutdown(self):
        "Sets self.stop to True and calls shutdown on tasks"
        self._stop = True
        if not self.rx is None:
            self.rx.shutdown()
        if not self.tx is None:
            self.tx.shutdown()
        if not self.server is None:
            self.server.close()


    async def __call__(self, readerque, writerque):
        "Called from indipydriver.asyncrun() to run the communications"
        self.readerque = readerque
        self.writerque = writerque
        logger.info(f"Listening on {self.host} : {self.port}")
        self.server = await asyncio.start_server(self.handle_data, self.host, self.port)
        try:
            async with self.server:
                await self.server.serve_forever()
        except asyncio.CancelledError:
            # self._stop raises an unwanted CancelledError
            # propogate this only if it is not due to self._stop
            if not self._stop:
                raise

    async def handle_data(self, reader, writer):
        "Used by asyncio.start_server, called to handle a client connection"
        if self.connected:
            # already connected, can only handle one connection
            writer.close()
            await writer.wait_closed()
            return
        self.connected = True
        addr = writer.get_extra_info('peername')
        self.rx = Port_RX(self.sendchecker, reader)
        self.tx = Port_TX(self.sendchecker, writer)
        logger.info(f"Connection received from {addr}")
        try:
            txtask = asyncio.create_task(self.tx.run_tx(self.writerque))
            rxtask = asyncio.create_task(self.rx.run_rx(self.readerque))
            await asyncio.gather(txtask, rxtask)
        except Exception as e:
            self.connected = False
            txtask.cancel()
            rxtask.cancel()
        finally:
            self.connected = False
            cleanque(self.writerque)
            logger.info(f"Connection from {addr} closed")


def cleanque(que):
    "Clears out a que"
    try:
        while True:
            xmldata = que.get_nowait()
            que.task_done()
    except asyncio.QueueEmpty:
        # que is now empty, nothing further to do
        pass


# Command to control whether setBLOBs should be sent to this channel from a given Device. They can
# be turned off completely by setting Never (the default), allowed to be intermixed with other INDI
# commands by setting Also or made the only command by setting Only.

# <!ELEMENT enableBLOB %BLOBenable >
# <!ATTLIST enableBLOB
# device %nameValue; #REQUIRED  name of Device
# name %nameValue; #IMPLIED name of BLOB Property, or all if absent
# >



class SendChecker:
    """Carries the enableBLOB status on a device, and does checks
       to ensure valid data is being transmitted"""

    def __init__(self, devices, exdrivers=None, remotes=None):
        "For every device create a dictionary"
        if remotes is None:
            self.remotes = []
        else:
            self.remotes = remotes
        if exdrivers is None:
            self.exdrivers = []
        else:
            self.exdrivers = exdrivers
        self.devices = devices
        self.devicestatus = {}
        # create a dictionary of devicenames :
        for devicename in devices:
            self.devicestatus[devicename] = {"Default":"Never", "Properties":{}}
            # The Properties value is a dictionary of propertyname:status

    def allowed(self, xmldata):
        "Return True if this xmldata can be transmitted, False otherwise"
        if xmldata.tag.startswith("new"):
            # new tags are sent from client to server, not from server back to client
            return False
        # allow anything with zero contents, such as getProperties
        if not len(xmldata):
            return True
        devicename = xmldata.get("device")
        if devicename is None:
            # enableBLOB only appliesto a specified device, not applicable here
            return True
        if not (devicename in self.devicestatus):
            # devicename not recognised, add it
            self.devicestatus[devicename] = {"Default":"Never", "Properties":{}}

        devicedict = self.devicestatus[devicename]

        # so we have a devicename, get propertyname
        name = xmldata.get("name")
        # if name missing, could be a message, cannot be a setBLOBVector
        if name is None:
            # If any property of this device has 'Only' set, then do not transmit
            if devicedict["Default"] == "Only":
                return False
            for value in devicedict["Properties"].values():  # gets status of each property
                if value == "Only":
                    return False
            return True

        # so we have a devicename, property name, is this xml a setBLOBVector
        if xmldata.tag == "setBLOBVector":
            if name in devicedict["Properties"]:
                if devicedict["Properties"][name] == "Never":
                    return False
                else:
                    return True
            elif devicedict["Default"] == "Never":
                return False
            else:
                return True

        # so not a setBLOBVector
        # If any property of this device has 'Only' set, then do not transmit
        if devicedict["Default"] == "Only":
            return False
        for value in devicedict["Properties"].values():  # gets status of each property
            if value == "Only":
                return False
        return True


    def setpermissions(self, rxdata):
        "Read the received enableBLOB xml and set permission in self.devicestatus"
        devicename = rxdata.get("device")
        if devicename is None:
            # invalid
            return
        if devicename not in self.devicestatus:
            # devicename not recognised, add it
            devicefound = False
            if (devicename in self.devices):
                devicefound = True
            if not devicefound:
                for exd in self.exdrivers:
                    if devicename in exd:
                        devicefound = True
                        break
            if not devicefound:
                for remcon in self.remotes:
                    if devicename in remcon:
                        devicefound = True
                        break
            if devicefound:
                self.devicestatus[devicename] = {"Default":"Never", "Properties":{}}
            else:
                # unknown device
                return


        # get the status of Never, Also, Only
        status = rxdata.text.strip()
        if status not in ("Never", "Also", "Only"):
            # invalid
            return

        devicedict = self.devicestatus[devicename]

        # property name
        name = rxdata.get("name")
        if name is None:
            # This applies to the device rather than to a particular property
            devicedict["Default"] = status
            return

        if name in devicedict["Properties"]:
            devicedict["Properties"][name] = status
            return

        # So this applies to a property that is not in self.devicestatus
        # check property is known, and add it
        propertyobject = None
        if devicename in self.devices:
            if name in self.devices[devicename]:
                propertyobject = self.devices[devicename][name]
            else:
                # devicename is in self.devices, but property not found
                return

        if propertyobject is None:
            for exd in self.exdrivers:
                if devicename in exd:
                    if name in exd.devicenames[devicename]:
                        propertyobject = exd.devicenames[devicename][name]
                        break
                    else:
                        # devicename in exd, but name not in exd.devicenames[devicename]
                        return

        if propertyobject is None:
            for remcon in self.remotes:
                if devicename in remcon:
                    if name in remcon[devicename]:
                        propertyobject = remcon[devicename][name]
                        break
                    else:
                        # devicename in remcon, but name not in remcon[devicename]
                        return

        if propertyobject is None:
            # property not known about, reject this
            return

        # confirm propertyobject is a BLOBVector
        if propertyobject.vectortype != "BLOBVector":
            return

        # add it to devicedict, and hence to self.devicestatus
        devicedict["Properties"][name] = status





import os, sys, collections, asyncio, time, copy, json

from time import sleep

from datetime import datetime, timezone

import xml.etree.ElementTree as ET

import logging
logger = logging.getLogger(__name__)

from . import events

from .propertymembers import ParseException


# All xml data received from the driver should be contained in one of the following tags
TAGS = (b'message',
        b'delProperty',
        b'defSwitchVector',
        b'setSwitchVector',
        b'defLightVector',
        b'setLightVector',
        b'defTextVector',
        b'setTextVector',
        b'defNumberVector',
        b'setNumberVector',
        b'defBLOBVector',
        b'setBLOBVector',
        b'getProperties'       # for snooping
       )

DEFTAGS = ( 'defSwitchVector',
            'defLightVector',
            'defTextVector',
            'defNumberVector',
            'defBLOBVector'
          )




# _STARTTAGS is a tuple of ( b'<defTextVector', ...  ) data received will be tested to start with such a starttag
_STARTTAGS = tuple(b'<' + tag for tag in TAGS)

# _ENDTAGS is a tuple of ( b'</defTextVector>', ...  ) data received will be tested to end with such an endtag
_ENDTAGS = tuple(b'</' + tag + b'>' for tag in TAGS)



def _makestart(element):
    "Given an xml element, returns a string of its start, including < tag attributes >"
    attriblist = ["<", element.tag]
    for key,value in element.attrib.items():
        attriblist.append(f" {key}=\"{value}\"")
    attriblist.append(">")
    return "".join(attriblist)



class IPyClient(collections.UserDict):

    """This class can be used to create your own scripts or client, and provides
       a connection to an INDI service, with parsing of the XML protocol.
       You should create your own class, inheriting from this, and overriding the
       rxevent method.
       The argument clientdata provides any named arguments you may wish to pass
       into the object when instantiating it.
       The IPyClient object is also a mapping of devicename to device object, which
       is populated as devices and their vectors are learned from the INDI protocol."""


    def __init__(self, indihost="localhost", indiport=7624, **clientdata):
        "An instance of this is a mapping of devicename to device object"
        super().__init__()

        # The UserDict will create self.data which will become
        # a dictionary of devicename to device object

        self.indihost = indihost
        self.indiport = indiport

        # dictionary of optional data
        self.clientdata = clientdata

        # create queue where client will put xml data to be transmitted
        self._writerque = asyncio.Queue(4)

        # and create readerque where received xmldata will be put
        self._readerque = asyncio.Queue(4)

        # self.messages is a deque of (Timestamp, message) tuples
        self.messages = collections.deque(maxlen=8)

        # note, messages are added with 'appendleft'
        # so newest message is messages[0]
        # oldest message is messages[-1] or can be obtained with .pop()

        # self.connected is True if connection has been made
        self.connected = False

        #####################
        # The following sets of timers are only enabled if this is True
        self.timeout_enable = True

        # vector timeouts are used to check that when a new vector is sent
        # a reply setvector will be received within the given time
        self.vector_timeout_min = 2
        self.vector_timeout_max = 10

        # idle_timer is set when either data is transmitted or received.
        # If nothing is sent or received after idle_timeout reached, then a getProperties is transmitted
        self.idle_timer = time.time()
        self.idle_timeout = 20
        # self.idle_timeout is set to two times self.vector_timeout_max

        # tx_timer is set when any data is transmitted,
        # it is used to check when any data is received,
        # at which point it becomes None again.
        # if there is no answer after self.respond_timeout seconds,
        # assume the connection has failed and close the connection
        self.tx_timer = None
        self.respond_timeout = 40
        # self.respond_timeout is set to four times self.vector_timeout_max
        ######################

        # and shutdown routine sets this to True to stop coroutines
        self._stop = False
        # this is set when asyncrun is finished
        self.stopped = asyncio.Event()

        # Indicates how verbose the debug xml logs will be when created.
        self._verbose = 1

        # Enables reports, by adding INFO logs to client messages
        self.enable_reports = True


    def debug_verbosity(self, verbose):
        """Set how verbose the debug xml logs will be when created.

           |  0 no xml logs will be generated
           |  1 for transmitted/received vector tags only,
           |  2 for transmitted/received vectors, members and contents (apart from BLOBs)
           |  3 for all transmitted/received data including BLOBs."""
        if verbose not in (0,1,2,3):
            raise ValueError
        self._verbose = verbose

    async def hardware(self):
        """This is started when asyncrun is called. As default does nothing so stops immediately.
           It is available to be overriden if required."""
        pass

    def shutdown(self):
        "Shuts down the client, sets the flag self._stop to True"
        self._stop = True

    @property
    def stop(self):
        "returns self._stop, being the instruction to stop the client"
        return self._stop


    async def queueput(self, queue, value, timeout=0.5):
        """Given an asyncio.Queue object, if self.stop is not set, this
           attempts to put value into the queue.
           If the queue is full, and the put operation is waiting, then
           after the timeout period the check and put will be repeated
           until successful, or self.stop becomes True.
           Returns True if value added to queue.
           Returns False if stop is True and the value not added."""
        while not self._stop:
            try:
                await asyncio.wait_for(queue.put(value), timeout)
            except asyncio.TimeoutError:
                # queue is full, continue while loop, checking stop flag
                continue
            return True
        return False


    async def report(self, message):
        """If logging is enabled message will be logged at level INFO.
           If self.enable_reports is True, the message will be injected into
           the received data, which will be picked up by the rxevent method.
           It is a way to set a message on to your client display, in the
           same way messages come from the INDI service."""
        try:
            logger.info(message)
            if not self.enable_reports:
                return
            timestamp = datetime.now(tz=timezone.utc)
            timestamp = timestamp.replace(tzinfo=None)
            root = ET.fromstring(f"<message timestamp=\"{timestamp.isoformat(sep='T')}\" message=\"{message}\" />")
            # and place root into readerque
            await self.queueput(self._readerque, root)
        except Exception :
            logger.exception("Exception report from IPyClient.report method")


    def enabledlen(self):
        "Returns the number of enabled devices"
        return sum(map(lambda x:1 if x.enable else 0, self.data.values()))


    def __setitem__(self, device):
        "Devices are added by being learnt from the driver, they cannot be manually added"
        raise KeyError


    async def _comms(self):
        "Create a connection to an INDI port"
        try:
            while not self._stop:
                self.tx_timer = None
                self.idle_timer = time.time()
                t1 = None
                t2 = None
                t3 = None
                try:
                    # start by openning a connection
                    await self.report(f"Attempting to connect to {self.indihost}:{self.indiport}")
                    reader, writer = await asyncio.open_connection(self.indihost, self.indiport)
                    self.connected = True
                    self.messages.clear()
                    # clear devices etc
                    self.clear()
                    await self.report(f"Connected to {self.indihost}:{self.indiport}")
                    t1 = asyncio.create_task(self._run_tx(writer))
                    t2 = asyncio.create_task(self._run_rx(reader))
                    t3 = asyncio.create_task(self._check_alive(writer))
                    await asyncio.gather(t1, t2, t3)
                except ConnectionRefusedError:
                    await self.report(f"Connection refused on {self.indihost}:{self.indiport}")
                except ConnectionError:
                    await self.report(f"Connection Lost on {self.indihost}:{self.indiport}")
                except Exception:
                    logger.exception(f"Connection Error on {self.indihost}:{self.indiport}")
                    await self.report("Connection failed")
                self._clear_connection()
                # connection has failed, ensure all tasks are done
                if t1:
                    while not t1.done():
                        await asyncio.sleep(0)
                if t2:
                    while not t2.done():
                        await asyncio.sleep(0)
                if t3:
                    while not t3.done():
                        await asyncio.sleep(0)
                if self._stop:
                    break
                else:
                    await self.report(f"Connection failed, re-trying...")
                # wait five seconds before re-trying, but keep checking
                # that self._stop has not been set
                count = 0
                while not self._stop:
                    await asyncio.sleep(0.5)
                    count += 1
                    if count >= 10:
                        break
        except Exception:
            logger.exception("Exception report from IPyClient._comms method")
            raise
        finally:
            self.shutdown()


    def _clear_connection(self):
        "On a connection closing down, self.connected is set to False"
        self.connected = False
        self.tx_timer = None



    async def send(self, xmldata):
        """Transmits xmldata, this is an internal method, not normally called by a user.
           xmldata is an xml.etree.ElementTree object"""
        if self.connected and (not self._stop):
            await self.queueput(self._writerque, xmldata)


    async def _check_alive(self, writer):
        try:
            while self.connected and (not self._stop):
                await asyncio.sleep(0.1)
                if self.tx_timer:
                    # data has been sent, waiting for reply
                    telapsed = time.time() - self.tx_timer
                    if telapsed > self.respond_timeout:
                        # no response to transmission self.respond_timeout seconds ago
                       writer.close()
                       await writer.wait_closed()
                       self._clear_connection()
                       if not self._stop:
                           await self.report("Error: Connection timed out")
            if self.connected and self._stop:
                writer.close()
                await writer.wait_closed()
                self._clear_connection()
        except Exception:
            logger.exception("Error in IPyClient._check_alive method")
            raise
        finally:
            self.connected = False

    def _logtx(self, txdata):
        "log tx data with level debug, and detail depends on self._verbose"
        if not self._verbose:
            return
        startlog = "TX:: "
        if self._verbose == 3:
            binarydata = ET.tostring(txdata)
            logger.debug(startlog + binarydata.decode())
        elif self._verbose == 2:
            if txdata.tag == "newBLOBVector" or txdata.tag == "setBLOBVector":
                data = copy.deepcopy(txdata)
                for element in data:
                    element.text = "NOT LOGGED"
                binarydata = ET.tostring(data)
            else:
                binarydata = ET.tostring(txdata)
            logger.debug(startlog + binarydata.decode())
        elif self._verbose == 1:
            data = copy.deepcopy(txdata)
            for element in data:
                data.remove(element)
            data.text = ""
            binarydata = ET.tostring(data, short_empty_elements=False).split(b">")
            logger.debug(startlog + binarydata[0].decode()+">")


    async def _run_tx(self, writer):
        "Monitors self._writerque and if it has data, uses writer to send it"
        try:
            while self.connected and (not self._stop):
                try:
                    txdata = await asyncio.wait_for(self._writerque.get(), timeout=0.5)
                except asyncio.TimeoutError:
                    continue
                self._writerque.task_done()
                if not self.connected:
                    break
                if self._stop:
                    break
                # send it out on the port
                binarydata = ET.tostring(txdata)
                # Send to the port
                writer.write(binarydata)
                await writer.drain()
                if self.timeout_enable:
                    # data has been transmitted set timers going, do not set timer
                    # for enableBLOB as no answer is expected for that
                    if (self.tx_timer is None) and (txdata.tag != "enableBLOB"):
                        self.tx_timer = time.time()
                self.idle_timer = time.time()
                if logger.isEnabledFor(logging.DEBUG):
                    self._logtx(txdata)
        except Exception:
            logger.exception("Exception report from IPyClient._run_tx method")
            raise

    def _logrx(self, rxdata):
        "log rx data to file"
        if not self._verbose:
            return
        startlog = "RX:: "
        if self._verbose == 3:
            binarydata = ET.tostring(rxdata)
            logger.debug(startlog + binarydata.decode())
        elif self._verbose == 2:
            data = copy.deepcopy(rxdata)
            tag = data.tag
            for element in data:
                if tag  == "newBLOBVector":
                    element.text = "NOT LOGGED"
            binarydata = ET.tostring(data)
            logger.debug(startlog + binarydata.decode())
        elif self._verbose == 1:
            data = copy.deepcopy(rxdata)
            for element in data:
                data.remove(element)
            data.text = ""
            binarydata = ET.tostring(data, short_empty_elements=False).split(b">")
            logger.debug(startlog + binarydata[0].decode() + ">")

    async def _run_rx(self, reader):
        "pass xml.etree.ElementTree data to readerque"
        try:
            # get block of xml.etree.ElementTree data
            # from self._xmlinput and append it to  readerque
            while self.connected and (not self._stop):
                rxdata = await self._xmlinput(reader)
                if rxdata is None:
                    return
                # and place rxdata into readerque
                result = await self.queueput(self._readerque, rxdata, 0.2)
                if not result:
                    # self._stop must be set
                    return
                # rxdata in readerque, log it, then continue with next block
                if logger.isEnabledFor(logging.DEBUG):
                    self._logrx(rxdata)
        except Exception:
            logger.exception("Exception report from IPyClient._run_rx")
            raise


    async def _xmlinput(self, reader):
        """get received data, parse it, and return it as xml.etree.ElementTree object
           Returns None if notconnected/stop flags arises"""
        message = b''
        messagetagnumber = None
        while self.connected and (not self._stop):
            await asyncio.sleep(0)
            data = await self._datainput(reader)
            # data is either None, or binary data ending in b">"
            if data is None:
                return
            if not self.connected:
                return
            if self._stop:
                return
            if not message:
                # data is expected to start with <tag, first strip any newlines
                data = data.strip()
                for index, st in enumerate(_STARTTAGS):
                    if data.startswith(st):
                        messagetagnumber = index
                        break
                    elif st in data:
                        # remove any data prior to a starttag
                        positionofst = data.index(st)
                        data = data[positionofst:]
                        messagetagnumber = index
                        break
                else:
                    # data does not start with a recognised tag, so ignore it
                    # and continue waiting for a valid message start
                    continue
                # set this data into the received message
                message = data
                # either further children of this tag are coming, or maybe its a single tag ending in "/>"
                if message.endswith(b'/>'):
                    # the message is complete, handle message here
                    try:
                        root = ET.fromstring(message.decode("us-ascii"))
                    except ET.ParseError as e:
                       # failed to parse the message, continue at beginning
                        message = b''
                        messagetagnumber = None
                        continue
                    # xml datablock done, return it
                    return root
                # and read either the next message, or the children of this tag
                continue
            # To reach this point, the message is in progress, with a messagetagnumber set
            # keep adding the received data to message, until an endtag is reached
            message += data
            if message.endswith(_ENDTAGS[messagetagnumber]):
                # the message is complete, handle message here
                try:
                    root = ET.fromstring(message.decode("us-ascii"))
                except ET.ParseError as e:
                    # failed to parse the message, continue at beginning
                    message = b''
                    messagetagnumber = None
                    continue
                # xml datablock done, return it
                return root
            # so message is in progress, with a messagetagnumber set
            # but no valid endtag received yet, so continue the loop


    async def _datainput(self, reader):
        """Waits for binary string of data ending in > from the port
           Returns None if notconnected/stop flags arises"""
        binarydata = b""
        while self.connected and (not self._stop):
            await asyncio.sleep(0)
            try:
                data = await reader.readuntil(separator=b'>')
            except asyncio.LimitOverrunError:
                data = await reader.read(n=32000)
            except asyncio.IncompleteReadError:
                binarydata = b""
                continue
            if not data:
                await asyncio.sleep(0.01)
                continue
            # data received
            self.tx_timer = None
            self.idle_timer = time.time()
            if b">" in data:
                binarydata = binarydata + data
                return binarydata
            # data has content but no > found
            binarydata += data
            # could put a max value here to stop this increasing indefinetly


    async def _rxhandler(self):
        """Populates the events using data from self._readerque"""
        try:
            while not self._stop:
                # get block of data from the self._readerque
                try:
                    root = await asyncio.wait_for(self._readerque.get(), timeout=0.5)
                except asyncio.TimeoutError:
                    # nothing to read, continue while loop which re-checks the _stop flag
                    continue
                devicename = root.get("device")
                try:
                    if devicename is None:
                        if root.tag == "message":
                            # create event
                            event = events.Message(root, None, self)
                        elif root.tag == "getProperties":
                            # create event
                            event = events.getProperties(root, None, self)
                        else:
                            # if no devicename and not message or getProperties, do nothing
                            continue
                    elif devicename in self:
                        # device is known about
                        device = self[devicename]
                        event = device.rxvector(root)
                    elif root.tag == "getProperties":
                        # device is not known about, but this is a getProperties, so raise an event
                        event = events.getProperties(root, None, self)
                    elif root.tag in DEFTAGS:
                        # device not known, but a def is received
                        newdevice = Device(devicename, self)
                        event = newdevice.rxvector(root)
                        # no error has occurred, so add this device to self.data
                        self.data[devicename] = newdevice
                    else:
                        # device not known, not a def or getProperties, so ignore it
                        continue
                except ParseException as pe:
                    # if a ParseException is raised, it is because received data is malformed
                    await self.report(str(pe))
                    continue
                finally:
                    self._readerque.task_done()
                # and to get here, continue has not been called
                # and an event has been created, call the user event handling function
                await self.rxevent(event)

        except Exception:
            logger.exception("Exception report from IPyClient._rxhandler method")
            raise
        finally:
            self.shutdown()


    def snapshot(self):
        """Take a snapshot of the client and returns an object which is a restricted copy
           of the current state of devices and vectors.
           Vector methods for sending data will not be available.
           These copies will not be updated by events. This is provided so that you can
           handle the client data, without fear of their values changing."""

        snap = Snap(self.indihost, self.indiport, self.connected, self.messages)
        if self.data:
            for devicename, device in self.data.items():
                snap[devicename] = device.snapshot()
        # return the snapshot
        return snap


    async def send_newVector(self, devicename, vectorname, timestamp=None, members={}):
        """Send a Vector with updated member values, members is a membername
           to value dictionary. Note, if this vector is a BLOB Vector, the members
           dictionary should be {membername:(value, blobsize, blobformat)}"""
        device = self.data.get(devicename)
        if device is None:
            return
        propertyvector = device.get(vectorname)
        if propertyvector is None:
            return
        try:
            if propertyvector.vectortype == "SwitchVector":
                await propertyvector.send_newSwitchVector(timestamp, members)
            elif propertyvector.vectortype == "TextVector":
                await propertyvector.send_newTextVector(timestamp, members)
            elif propertyvector.vectortype == "NumberVector":
                await propertyvector.send_newNumberVector(timestamp, members)
            elif propertyvector.vectortype == "BLOBVector":
                await propertyvector.send_newBLOBVector(timestamp, members)
        except Exception:
            logger.exception("Exception report from IPyClient.send_newVector method")
            raise


    def set_vector_timeouts(self, timeout_enable=None, timeout_min=None, timeout_max=None):
        """The INDI protocol allows the server to suggest a timeout for each vector. This
           method allows you to set minimum and maximum timeouts which restricts the
           suggested values.

           These should be given as integer seconds. If any parameter
           is not provided (left at None) then that value will not be changed.

           If timeout_enable is set to False, no VectorTimeOut events will occur.

           As default, timeouts are enabled, minimum is set to 2 seconds, maximum 10 seconds.
           """
        if not timeout_enable is None:
            self.timeout_enable = timeout_enable
        if not timeout_min is None:
            self.vector_timeout_min = timeout_min
        if not timeout_max is None:
            self.vector_timeout_max = timeout_max
            self.idle_timeout = 2 * timeout_max
            self.respond_timeout = 4 * timeout_max


    async def _timeout_monitor(self):
        """Sends a getProperties every five seconds if no devices have been learnt
           or every self.idle_timeout seconds if nothing has been transmitted or received"""
        try:
            count = 0
            while (not self._stop):
                await asyncio.sleep(0.5)
                # This loop tests timeout values every half second
                if not self.connected:
                    count = 0
                else:
                    # so the connection is up, check enabled devices exist
                    if self.enabledlen():
                        # connection is up and devices exist
                        if self.timeout_enable:
                            # If nothing has been sent or received
                            # for self.idle_timeout seconds, send a getProperties
                            nowtime = time.time()
                            telapsed = nowtime - self.idle_timer
                            if telapsed > self.idle_timeout:
                                await self.send_getProperties()
                            # check if any vectors have timed out
                            for device in self.data.values():
                                for vector in device.values():
                                    if not vector.enable:
                                        continue
                                    if vector.checktimedout(nowtime):
                                        # Creat a VectorTimeOut event
                                        event = events.VectorTimeOut(device, vector)
                                        await self.rxevent(event)
                    else:
                        # no devices
                        # then send a getProperties, every five seconds, when count is zero
                        if not count:
                            await self.send_getProperties()
                            await self.report("getProperties sent")
                        count += 1
                        if count >= 10:
                            count = 0
        except Exception:
            logger.exception("Exception report from IPyClient._timeout_monitor method")
            raise
        finally:
            self.shutdown()


    async def send_getProperties(self, devicename=None, vectorname=None):
        """Sends a getProperties request. On startup the IPyClient object
           will automatically send getProperties, so typically you will
           not have to use this method."""
        if self.connected:
            xmldata = ET.Element('getProperties')
            xmldata.set("version", "1.7")
            if not devicename:
                await self.send(xmldata)
                return
            xmldata.set("device", devicename)
            if vectorname:
                xmldata.set("name", vectorname)
            await self.send(xmldata)

    async def send_enableBLOB(self, value, devicename, vectorname=None):
        """Sends an enableBLOB instruction. The value should be one of "Never", "Also", "Only"."""
        if self.connected:
            if value not in ("Never", "Also", "Only"):
                return
            xmldata = ET.Element('enableBLOB')
            if not devicename:
                # a devicename is required
                return
            xmldata.set("device", devicename)
            if vectorname:
                xmldata.set("name", vectorname)
            xmldata.text = value
            await self.send(xmldata)

    def get_vector_state(self, devicename, vectorname):
        """Gets the state string of the given vectorname, if this vector does not exist
           returns None - this could be because the vector is not yet learnt.
           The vector state attribute will still be returned, even if vector.enable is False"""
        device = self.data.get(devicename)
        if device is None:
            return
        propertyvector = device.get(vectorname)
        if propertyvector is None:
            return
        return propertyvector.state


    async def rxevent(self, event):
        """Override this.
           On receiving data, this is called, and should handle any necessary actions.
           event is an object with attributes according to the data received."""
        pass


    async def asyncrun(self):
        "Await this method to run the client."
        self._stop = False
        try:
            await asyncio.gather(self._comms(), self._rxhandler(), self._timeout_monitor(), self.hardware())
        finally:
            self.stopped.set()
            self._stop = True



class Snap(collections.UserDict):

    """An instance of this object is returned when a snapshot is
       taken of the client.
       It is a mapping of device name to device snapshots, which
       are in turn mappings of vectorname to vector snapshots.
       These snapshots record values and attributes, at the
       moment of the snapshot.
       Unlike IPyClient this has no send_newVector method, and the
       snap vectors do not have the send methods."""

    def __init__(self, indihost, indiport, connected, messages):
        super().__init__()
        self.indihost = indihost
        self.indiport = indiport
        self.connected = connected
        self.messages = list(messages)


    def enabledlen(self):
        "Returns the number of enabled devices"
        return sum(map(lambda x:1 if x.enable else 0, self.data.values()))


    def get_vector_state(self, devicename, vectorname):
        """Gets the state string of the given vectorname, if this vector does not exist
           returns None"""
        device = self.data.get(devicename)
        if device is None:
            return
        propertyvector = device.get(vectorname)
        if propertyvector is None:
            return
        return propertyvector.state


    def dictdump(self):
        """Returns a dictionary of this client information
           and is used to generate the JSON output"""
        messlist = []
        for message in self.messages:
            messlist.append([message[0].isoformat(sep='T'), message[1]])
        devdict = {}
        for devicename, device in self.items():
            devdict[devicename] = device.dictdump()
        return {"indihost":self.indihost,
                "indiport":self.indiport,
                "connected":self.connected,
                "messages":messlist,
                "enable":self.enable,
                "devices":devdict}

    def dumps(self, indent=None, separators=None):
        "Returns a JSON string of the snapshot."
        return json.dumps(self.dictdump(), indent=indent, separators=separators)


    def dump(self, fp, indent=None, separators=None):
        """Serialize the snapshot as a JSON formatted stream to fp, a file-like object.
           This uses the Python json module which always produces str objects, not bytes
           objects. Therefore, fp.write() must support str input."""
        return json.dump(self.dictdump(), fp, indent=indent, separators=separators)



class _ParentDevice(collections.UserDict):
    "Each device is a mapping of vector name to vector object."

    def __init__(self, devicename):
        super().__init__()
        # self.data is created by UserDict and will become a
        # dictionary of vector name to vector this device owns

        # This device name
        self.devicename = devicename


    @property
    def enable(self):
        "Returns True if any vector of this device has enable True, otherwise False"
        for vector in self.data.values():
            if vector.enable:
                return True
        return False

    def disable(self):
        "If called, disables the device"
        for vector in self.data.values():
            vector.enable = False


class SnapDevice(_ParentDevice):
    """This object is used as a snapshot of this device
       It is a mapping of vector name to vector snapshots"""

    def __init__(self, devicename, messages):
        super().__init__(devicename)
        self.messages = list(messages)

    def dictdump(self):
        """Returns a dictionary of this device information
           and is used to generate the JSON output"""
        messlist = []
        for message in self.messages:
            messlist.append([message[0].isoformat(sep='T'), message[1]])
        vecdict = {}
        for vectorname, vector in self.items():
            vecdict[vectorname] = vector.dictdump()
        return {"devicename":self.devicename, "messages":messlist, "enable":self.enable, "vectors":vecdict}

    def dumps(self, indent=None, separators=None):
        "Returns a JSON string of the snapshot."
        return json.dumps(self.dictdump(), indent=indent, separators=separators)


    def dump(self, fp, indent=None, separators=None):
        """Serialize the snapshot as a JSON formatted stream to fp, a file-like object.
           This uses the Python json module which always produces str objects, not bytes
           objects. Therefore, fp.write() must support str input."""
        return json.dump(self.dictdump(), fp, indent=indent, separators=separators)



class Device(_ParentDevice):

    """An instance of this is created for each device
       as data is received.
    """

    def __init__(self, devicename, client):
        super().__init__(devicename)

        # and the device has a reference to its client
        self._client = client

        # self.messages is a deque of tuples (timestamp, message)
        self.messages = collections.deque(maxlen=8)


    def __setitem__(self, propertyname, propertyvector):
        "Properties are added by being learnt from the driver, they cannot be manually added"
        raise KeyError


    def rxvector(self, root):
        """Handle received data, sets new propertyvector into self.data,
           or updates existing property vector and returns an event"""
        try:
            if root.tag == "delProperty":
                return events.delProperty(root, self, self._client)
            elif root.tag == "message":
                return events.Message(root, self, self._client)
            elif root.tag == "defSwitchVector":
                return events.defSwitchVector(root, self, self._client)
            elif root.tag == "setSwitchVector":
                return events.setSwitchVector(root, self, self._client)
            elif root.tag == "defLightVector":
                return events.defLightVector(root, self, self._client)
            elif root.tag == "setLightVector":
                return events.setLightVector(root, self, self._client)
            elif root.tag == "defTextVector":
                return events.defTextVector(root, self, self._client)
            elif root.tag == "setTextVector":
                return events.setTextVector(root, self, self._client)
            elif root.tag == "defNumberVector":
                return events.defNumberVector(root, self, self._client)
            elif root.tag == "setNumberVector":
                return events.setNumberVector(root, self, self._client)
            elif root.tag == "defBLOBVector":
                return events.defBLOBVector(root, self, self._client)
            elif root.tag == "setBLOBVector":
                return events.setBLOBVector(root, self, self._client)
            elif root.tag == "getProperties":
                return events.getProperties(root, self, self._client)
            else:
                raise ParseException("Unrecognised tag received")
        except ParseException:
            raise
        except Exception:
            logger.exception("Exception report from IPyClient.rxvector method")
            raise ParseException("Error while attempting to parse received data")


    def snapshot(self):
        """Take a snapshot of the device and returns an object which is a restricted copy
           of the current state of the device and its vectors.
           Vector methods for sending data will not be available.
           This copy will not be updated by events. This is provided so that you can
           handle the device data, without fear of the value changing."""
        snapdevice = SnapDevice(self.devicename, self.messages)
        for vectorname, vector in self.data.items():
            snapdevice[vectorname] = vector.snapshot()
        return snapdevice




import sys

from datetime import datetime, timezone

from base64 import standard_b64decode

from collections import UserDict

import xml.etree.ElementTree as ET

from . import propertyvectors

from .propertymembers import ParseException, getfloat



def _parse_timestamp(timestamp_string):
    """Parse a timestamp string and return either None on failure, or a datetime object
       If the given timestamp_string is None, return the datetime for the current time.
       Everything is UTC"""
    if timestamp_string:
        try:
            if '.' in timestamp_string:
                # remove fractional part, not supported by datetime.fromisoformat
                timestamp_string, remainder = timestamp_string.rsplit('.', maxsplit=1)
                if len(remainder) < 6:
                    remainder = "{:<06}".format(remainder)
                elif len(remainder) > 6:
                    remainder = remainder[:6]
                remainder = int(remainder)
                timestamp = datetime.fromisoformat(timestamp_string)
                timestamp = timestamp.replace(microsecond=remainder, tzinfo=timezone.utc)
            else:
                timestamp = datetime.fromisoformat(timestamp_string)
                timestamp = timestamp.replace(tzinfo=timezone.utc)
        except Exception:
            timestamp = None
    else:
        timestamp = datetime.now(tz=timezone.utc)
    return timestamp


class VectorTimeOut:
    """This event is generated by a timeout, not by received data."""

    def __init__(self, device, vector):
        self.device = device
        self.devicename = self.device.devicename
        self.vector = vector
        self.vectorname = self.vector.name
        self.timestamp = datetime.now(tz=timezone.utc)
        self.eventtype = "TimeOut"
        self.root = None


class Event:
    "Parent class for events received from drivers"
    def __init__(self, root, device, client):
        self.device = device
        self._client = client
        self.vectorname = None
        if device is None:
            self.devicename = None
        else:
            self.devicename = self.device.devicename
        self.root = root
        self.timestamp = _parse_timestamp(root.get("timestamp"))

    def __str__(self):
        return ET.tostring(self.root, encoding='unicode')



class Message(Event):
    """This contains attribute 'message' with the message string sent by the remote driver.
       Attribute devicename could be None if the driver is sending a system wide message."""

    def __init__(self, root, device, client):
        super().__init__(root, device, client)
        self.eventtype = "Message"
        self.message = root.get("message", "")
        if device is None:
            # state wide message
            client.messages.appendleft( (self.timestamp, self.message) )
        else:
            device.messages.appendleft( (self.timestamp, self.message) )


class getProperties(Event):
    """This may have a device and name, both may be None
       But there may still be a devicename and vectorname for an unknown device"""

    def __init__(self, root, device, client):
        super().__init__(root, device, client)
        self.eventtype = "getProperties"
        self.devicename = root.get("device")
        self.vectorname = root.get("name")


class delProperty(Event):
    """The remote driver is instructing the client to delete either a device or a vector property.
       This contains attribute vectorname, if it is None, then the whole device is to be deleted.
       A 'message' attribute contains any message sent by the client with this instruction.
       This event will automatically set the appropriate enable flag to False in the effected
       device and vectors."""

    def __init__(self, root, device, client):
        super().__init__(root, device, client)
        self.eventtype = "Delete"
        if self.devicename is None:
            raise ParseException("delProperty has no devicename")
        if not self.device.enable:
            # already deleted
            raise ParseException("device already deleted")
        self.vectorname = root.get("name")
        self.message = root.get("message", "")
        # properties is a dictionary of property name to propertyvector this device owns
        # This method updates a property vector and sets it into properties
        properties = device.data
        if self.vectorname:
            # does this vector already exist, if it does, disable it
            if self.vectorname in properties:
                vector = properties[self.vectorname]
                vector.enable = False
        else:
            # No vectorname given, disable all properties
            for vector in properties.values():
                vector.enable = False


class defVector(Event, UserDict):
    "Parent to def vectors, adds a mapping of membername:value"
    def __init__(self, root, device, client):
        Event.__init__(self, root, device, client)
        UserDict.__init__(self)
        self.eventtype = "Define"
        self.vectorname = root.get("name")
        if self.vectorname is None:
            raise ParseException("defVector has no vector name")
        self.label = root.get("label", self.vectorname)
        self.group = root.get("group", "DEFAULT GROUP")
        state = root.get("state")
        if not state:
            raise ParseException("defVector has no state given")
        if not state in ('Idle','Ok','Busy','Alert'):
            raise ParseException("defVector has invalid state")
        self.state = state
        self.message = root.get("message", "")


    def __setitem__(self, membername):
        raise KeyError


class defSwitchVector(defVector):

    """The remote driver has sent this to define a switch vector property, this
       is a mapping of membername:value"""

    def __init__(self, root, device, client):
        defVector.__init__(self, root, device, client)
        self.perm = root.get("perm")
        if self.perm is None:
            raise ParseException("defSwitchVector has no perm given")
        if self.perm not in ('ro', 'wo', 'rw'):
            raise ParseException("defSwitchVector has invalid perm")
        self.rule = root.get("rule")
        if self.rule is None:
            raise ParseException("defSwitchVector has no rule given")
        if self.rule not in ('OneOfMany', 'AtMostOne', 'AnyOfMany'):
            raise ParseException("defSwitchVector has invalid rule")
        try:
            timeout = root.get("timeout")
            if not timeout:
                self.timeout = 0.0
            else:
                self.timeout = float(timeout)
        except Exception:
            self.timeout = 0.0
        # create object dictionary of member name to value
        # and another dictionary of self.memberlabels with key member name and value being label
        self.memberlabels = {}
        for member in root:
            if member.tag == "defSwitch":
                membername = member.get("name")
                if not membername:
                    raise ParseException("defSwitch member has no name")
                label = member.get("label", membername)
                self.memberlabels[membername] = label
                if not member.text:
                    raise ParseException("defSwitch member has invalid value")
                value = member.text.strip()
                if value == "On":
                    self.data[membername] = "On"
                elif value == "Off":
                    self.data[membername] = "Off"
                else:
                    raise ParseException("defSwitch member has invalid value")
            else:
                raise ParseException("defSwitchVector member has invalid tag")
        if not self.data:
            raise ParseException("defSwitchVector has no valid contents")

        # This method updates a property vector and sets it into the device 'data' dictionary
        # which is a dictionary of property name to propertyvector this device owns
        properties = device.data

        # does this vector already exist
        if self.vectorname in properties:
            self.vector = properties[self.vectorname]
            # set changed values into self.vector by calling vector._defvector
            # with this event as its argument
            self.vector._defvector(self)
        else:
            # create a new SwitchVector
            self.vector = propertyvectors.SwitchVector(self)
            # add it to properties
            properties[self.vectorname] = self.vector


class defTextVector(defVector):

    """The remote driver has sent this to define a text vector property, this
       is a mapping of membername:value."""

    def __init__(self, root, device, client):
        defVector.__init__(self, root, device, client)
        self.perm = root.get("perm")
        if self.perm is None:
            raise ParseException("No perm given in defTextVector")
        if self.perm not in ('ro', 'wo', 'rw'):
            raise ParseException("Invalid perm given in defTextVector")
        try:
            timeout = root.get("timeout")
            if not timeout:
                self.timeout = 0.0
            else:
                self.timeout = float(timeout)
        except Exception:
            self.timeout = 0.0
        # create object dictionary of member name to value
        # and another dictionary of self.memberlabels with key member name and value being label
        self.memberlabels = {}
        for member in root:
            if member.tag == "defText":
                membername = member.get("name")
                if not membername:
                    raise ParseException("Missing member name in defText")
                label = member.get("label", membername)
                self.memberlabels[membername] = label
                if not member.text:
                    value = ""
                else:
                    value = member.text.strip()
                self.data[membername] = value
            else:
                raise ParseException("Invalid tag in defTextVector")
        if not self.data:
            raise ParseException("No member values in defTextVector")

        # This method updates a property vector and sets it into the device 'data' dictionary
        # which is a dictionary of property name to propertyvector this device owns
        properties = device.data

        # does this vector already exist
        if self.vectorname in properties:
            self.vector = properties[self.vectorname]
            # set changed values into self.vector by calling vector._defvector
            # with this event as its argument
            self.vector._defvector(self)
        else:
            # create a new TextVector
            self.vector = propertyvectors.TextVector(self)
            # add it to properties
            properties[self.vectorname] = self.vector


class defNumberVector(defVector):

    """The remote driver has sent this to define a number vector property, this
       is a mapping of membername:value. Its attributes memberlabels gives further
       description of members, being a dictionary of
       membername:(label, format, min, max, step)."""

    def __init__(self, root, device, client):
        defVector.__init__(self, root, device, client)
        self.perm = root.get("perm")
        if self.perm is None:
            raise ParseException("No perm given in defNumberVector")
        if self.perm not in ('ro', 'wo', 'rw'):
            raise ParseException("Invalid perm given in defNumberVector")
        try:
            timeout = root.get("timeout")
            if not timeout:
                self.timeout = 0.0
            else:
                self.timeout = float(timeout)
        except Exception:
            self.timeout = 0.0
        # create object dictionary of member name to value
        # and another dictionary of self.memberlabels with key member name and
        # value being a tuple of (label, format, min, max, step)
        self.memberlabels = {}
        for member in root:
            if member.tag == "defNumber":
                membername = member.get("name")
                if not membername:
                    raise ParseException("Missing member name in defNumber")
                label = member.get("label", membername)
                memberformat = member.get("format")
                if not memberformat:
                    raise ParseException("Missing format in defNumber")
                membermin = member.get("min")
                if not membermin:
                    raise ParseException("Missing min value in defNumber")
                membermax = member.get("max")
                if not membermax:
                    raise ParseException("Missing max value in defNumber")
                memberstep = member.get("step")
                if not memberstep:
                    raise ParseException("Missing step value in defNumber")
                self.memberlabels[membername] = (label, memberformat, membermin, membermax, memberstep)
                if not member.text:
                    raise ParseException("Missing content in defNumber")
                self.data[membername] = member.text.strip()
            else:
                raise ParseException("Invalid tag in defNumberVector")
        if not self.data:
            raise ParseException("No member values in defNumberVector")


        # This method updates a property vector and sets it into the device 'data' dictionary
        # which is a dictionary of property name to propertyvector this device owns
        properties = device.data

        # does this vector already exist
        if self.vectorname in properties:
            self.vector = properties[self.vectorname]
            # set changed values into self.vector
            self.vector._defvector(self)
        else:
            # create a new NumberVector
            self.vector = propertyvectors.NumberVector(self)
            # add it to properties
            properties[self.vectorname] = self.vector



class defLightVector(defVector):

    """The remote driver has sent this to define a light vector property.
       This is a mapping of membername:value."""

    def __init__(self, root, device, client):
        defVector.__init__(self, root, device, client)
        # create object dictionary of member name to value
        # and another dictionary of self.memberlabels with key member name and value being label
        self.memberlabels = {}
        for member in root:
            if member.tag == "defLight":
                membername = member.get("name")
                if not membername:
                    raise ParseException("No name given in defLight")
                label = member.get("label", membername)
                self.memberlabels[membername] = label
                if not member.text:
                    raise ParseException("No value given in defLight")
                value = member.text.strip()
                if not value in ('Idle','Ok','Busy','Alert'):
                    raise ParseException("Invalid value given in defLight")
                self.data[membername] = value
            else:
                raise ParseException("No members of defLight present")
        if not self.data:
            raise ParseException("No member values present")


        # This method updates a property vector and sets it into the device 'data' dictionary
        # which is a dictionary of property name to propertyvector this device owns
        properties = device.data

        # does this vector already exist
        if self.vectorname in properties:
            self.vector = properties[self.vectorname]
            # set changed values into self.vector by calling vector._defvector
            # with this event as its argument
            self.vector._defvector(self)
        else:
            # create a new LightVector
            self.vector = propertyvectors.LightVector(self)
            # add it to properties
            properties[self.vectorname] = self.vector



class defBLOBVector(defVector):

    """The remote driver has sent this to define a BLOB vector property.

       However this class does not have an object mapping of member name to value, since
       values are not given in defBLOBVectors"""


    def __init__(self, root, device, client):
        defVector.__init__(self, root, device, client)
        # self.data created by UserDict is not populated
        self.eventtype = "DefineBLOB"
        if self.devicename is None:
            raise ParseException("No device name given in defBLOBVector")
        self.perm = root.get("perm")
        if self.perm is None:
            raise ParseException("No perm given in defBLOBVector")
        if self.perm not in ('ro', 'wo', 'rw'):
            raise ParseException("Invalid perm given in defBLOBVector")
        try:
            timeout = root.get("timeout")
            if not self.timeout:
                self.timeout = 0.0
            else:
                self.timeout = float(timeout)
        except Exception:
            self.timeout = 0.0
        # create a dictionary of self.memberlabels with key member name and value being label
        self.memberlabels = {}
        for member in root:
            if member.tag == "defBLOB":
                membername = member.get("name")
                if not membername:
                    raise ParseException("Missing member name in defBLOB")
                label = member.get("label", membername)
                self.memberlabels[membername] = label
            else:
                raise ParseException(f"Invalid child tag {member.tag} of defBLOBVector received")
        if not self.memberlabels:
            raise ParseException("No labels given in defBLOBVector")

        # properties is a dictionary of property name to propertyvector this device owns
        # This method updates a property vector and sets it into properties
        properties = device.data

        # does this vector already exist
        if self.vectorname in properties:
            self.vector = properties[self.vectorname]
            # set changed values into self.vector by calling vector._defvector
            # with this event as its argument
            self.vector._defvector(self)
        else:
            # create a new BLOBVector
            self.vector = propertyvectors.BLOBVector(self)
            # add it to properties
            properties[self.vectorname] = self.vector


class setVector(Event, UserDict):
    "Parent to set vectors, adds dictionary"
    def __init__(self, root, device, client):
        Event.__init__(self, root, device, client)
        UserDict.__init__(self)
        self.eventtype = "Set"
        if self.devicename is None:
            raise ParseException("Missing device name in set vector")
        self.vectorname = root.get("name")
        if self.vectorname is None:
            raise ParseException("Missing vector name in set vector")
        # This vector must already exist, and be enabled
        self.timeout = None
        if self.vectorname in self.device:
            vector = self.device[self.vectorname]
            if not vector.enable:
                raise ParseException("Set vector ignored as vector deleted")
        else:
            raise ParseException("Set vector ignored as vector is unknown")
        state = root.get("state")
        if state and (state in ('Idle','Ok','Busy','Alert')):
            self.state = state
        else:
            self.state = None
        self.message = root.get("message", "")

    def __setitem__(self, membername):
        raise KeyError



class setSwitchVector(setVector):
    """The remote driver is setting a Switch vector property.
       This is a mapping of membername:value."""

    def __init__(self, root, device, client):
        setVector.__init__(self, root, device, client)
        try:
            timeout = root.get("timeout")
            if not timeout is None:
                self.timeout = float(timeout)
        except Exception:
            # dont update
            pass
        # create a dictionary of member name to value
        for member in root:
            if member.tag == "oneSwitch":
                membername = member.get("name")
                if not membername:
                    raise ParseException("Missing name in oneSwitch")
                if not member.text:
                    raise ParseException("Missing value in oneSwitch")
                value = member.text.strip()
                if value == "On":
                    self.data[membername] = "On"
                elif value == "Off":
                    self.data[membername] = "Off"
                else:
                    raise ParseException("Invalid value in oneSwitch")
            else:
                raise ParseException("Invalid child tag of setSwitchVector")
        properties = device.data
        self.vector = properties[self.vectorname]
        # set changed values into self.vector
        self.vector._setvector(self)


class setTextVector(setVector):

    """The remote driver is setting a Text vector property.
       This is a mapping of membername:value."""

    def __init__(self, root, device, client):
        setVector.__init__(self, root, device, client)
        try:
            timeout = root.get("timeout")
            if not timeout is None:
                self.timeout = float(timeout)
        except Exception:
            # dont update
            pass
        # create a dictionary of member name to value
        for member in root:
            if member.tag == "oneText":
                membername = member.get("name")
                if not membername:
                    raise ParseException("Missing name in oneText")
                if not member.text:
                    value = ""
                else:
                    value = member.text.strip()
                self.data[membername] = value
            else:
                raise ParseException("Invalid child tag of setTextVector")
        properties = device.data
        self.vector = properties[self.vectorname]
        # set changed values into self.vector
        self.vector._setvector(self)


class setNumberVector(setVector):

    """The remote driver is setting a Number vector property.
       This is a mapping of membername:value.
       These number values are string values."""

    def __init__(self, root, device, client):
        setVector.__init__(self, root, device, client)
        try:
            timeout = root.get("timeout")
            if not timeout is None:
                self.timeout = float(timeout)
        except Exception:
            # dont update
            pass
        # create a dictionary of member name to value
        for member in root:
            if member.tag == "oneNumber":
                membername = member.get("name")
                if not membername:
                    raise ParseException("Missing name in oneNumber")
                if not member.text:
                    raise ParseException("Missing value in oneNumber")
                membervalue = member.text.strip()
                if not membervalue:
                    raise ParseException("Missing value in oneNumber")
                # test membervalue ok
                try:
                    memberfloat = getfloat(membervalue)
                except TypeError:
                    raise ParseException("Invalid number in setNumberVector")
                self.data[membername] = membervalue
            else:
                raise ParseException("Invalid child tag of setNumberVector")
        properties = device.data
        self.vector = properties[self.vectorname]
        # set changed values into self.vector
        self.vector._setvector(self)


class setLightVector(setVector):

    """The remote driver is setting a Light vector property.
       This is a mapping of membername:value.
       Note, the timeout attribute will always be None"""

    def __init__(self, root, device, client):
        setVector.__init__(self, root, device, client)
        # create a dictionary of member name to value
        for member in root:
            if member.tag == "oneLight":
                membername = member.get("name")
                if not membername:
                    raise ParseException("Missing name in oneLight")
                if not member.text:
                    raise ParseException("Missing value in oneLight")
                value = member.text.strip()
                if not value in ('Idle','Ok','Busy','Alert'):
                    raise ParseException("Invalid value in oneLight")
                self.data[membername] = value
            else:
                raise ParseException("Invalid child tag of setLightVector")
        properties = device.data
        self.vector = properties[self.vectorname]
        # set changed values into self.vector
        self.vector._setvector(self)


class setBLOBVector(setVector):

    """The remote driver is setting a BLOB vector property.
       This is a mapping of membername:value, where value is a
       bytes object, taken from the received xml and b64 decoded
       This event has further attribute sizeformat being a dictionary
       of membername:(size, format) and which are then set into the target
       members as blobsize and blobformat attributes."""


    def __init__(self, root, device, client):
        setVector.__init__(self, root, device, client)
        self.eventtype = "SetBLOB"
        try:
            timeout = root.get("timeout")
            if not timeout is None:
                self.timeout = float(timeout)
        except Exception:
            # dont update
            pass
        # create a dictionary of member name to value
        # and dictionary sizeformat
        # with key member name and value being a tuple of size, format
        self.sizeformat = {}
        for member in root:
            if member.tag == "oneBLOB":
                membername = member.get("name")
                if not membername:
                    raise ParseException("Missing name in oneBLOB")
                membersize = member.get("size")
                if not membersize:
                    raise ParseException("Missing size in oneBLOB")
                try:
                    memberize = int(membersize)
                except Exception:
                    raise ParseException("Invalid size in oneBLOB")
                memberformat = member.get("format")
                if not memberformat:
                    raise ParseException("Missing format in oneBLOB")
                if not member.text:
                    raise ParseException("Missing value in oneBLOB")
                try:
                    self.data[membername] = standard_b64decode(member.text.encode('ascii'))
                except Exception:
                    raise ParseException("Unable to decode oneBLOB contents")
                self.sizeformat[membername] = (membersize, memberformat)
            else:
                raise ParseException("Invalid child tag of setBLOBVector")
        self.vector = device[self.vectorname]
        # set changed values into self.vector
        self.vector._setvector(self)
